package db

import (
	"fmt"
	"io/ioutil"
	"math/rand"
	"os"
	"os/exec"
	"testing"
	"time"

	ippb "git.autistici.org/ai3/tools/iprep/proto"
	"github.com/syndtr/goleveldb/leveldb/util"
)

const manyEvents = 500000

func randomIP() string {
	return fmt.Sprintf("%d.%d.%d.%d", 1+rand.Intn(253), 1+rand.Intn(253), 1+rand.Intn(253), 1+rand.Intn(253))
}

func randomType() string {
	return "detection"
}

func randomEvent() *ippb.Event {
	return &ippb.Event{
		Ip:    randomIP(),
		Type:  randomType(),
		Count: 1,
	}
}

func randomEvents(n int) *ippb.Aggregate {
	aggr := new(ippb.Aggregate)
	for i := 0; i < n; i++ {
		aggr.AddEvent(randomEvent())
	}
	return aggr
}

func TestWrite(t *testing.T) {
	dir, err := ioutil.TempDir("", "")
	if err != nil {
		t.Fatal(err)
	}
	defer os.RemoveAll(dir)

	db, err := Open(dir + "/test.db")
	if err != nil {
		t.Fatalf("Open: %v", err)
	}
	defer db.Close()

	// Write n elements.
	n := manyEvents
	events := randomEvents(n)
	if err := db.AddAggregate(events); err != nil {
		t.Fatalf("AddEvents: %v", err)
	}

	data, _ := exec.Command("du", "-sh", dir).Output()
	t.Logf("added %d records: %s", n, string(data))

	// Scan events from 1 hour in the past, for the first IP in
	// events, should fetch a non-zero number of events.
	refType := events.ByType[0].Type
	refIP := events.ByType[0].ByIp[0].Ip
	m, err := db.ScanIP(time.Now().Add(-1*time.Hour), refIP)
	if err != nil {
		t.Fatalf("ScanIP(%s): %v", refIP, err)
	}
	if count := m[refType][refIP]; count < 1 {
		t.Fatalf("read %d events from db, expected > 0", count)
	}

	// Scan events from 1 hour in the future, should return 0 results.
	// count = 0
	// if err := db.Scan(time.Now().Add(1*time.Hour), func(*ippb.Aggregate) error {
	// 	return errors.New("unexpected Aggregate")
	// }, func(*ippb.Event) error {
	// 	count++
	// 	return nil
	// }); err != nil {
	// 	t.Fatalf("Scan: %v", err)
	// }
	// if count != 0 {
	// 	t.Fatalf("scan in the future read %d events from db, expected 0", count)
	// }
}

func TestWipeOldData(t *testing.T) {
	dir, err := ioutil.TempDir("", "")
	if err != nil {
		t.Fatal(err)
	}
	defer os.RemoveAll(dir)

	db, err := Open(dir + "/test.db")
	if err != nil {
		t.Fatalf("Open: %v", err)
	}
	defer db.Close()

	if err := db.AddAggregate(randomEvents(100)); err != nil {
		t.Fatalf("AddEvents: %v", err)
	}
	if err := db.WipeOldData(0); err != nil {
		t.Fatalf("WipeOldData: %v", err)
	}
}

func BenchmarkWrite(b *testing.B) {
	dir, err := ioutil.TempDir("", "")
	if err != nil {
		b.Fatal(err)
	}
	defer os.RemoveAll(dir)

	db, err := Open(dir + "/test.db")
	if err != nil {
		b.Fatalf("Open: %v", err)
	}
	defer db.Close()

	// Chunks of 100.
	chunkSize := 1000
	nchunks := b.N / chunkSize
	events := randomEvents(chunkSize)

	b.ResetTimer()

	for i := 0; i < nchunks; i++ {
		if err := db.AddAggregate(events); err != nil {
			b.Fatalf("AddEvents: %v", err)
		}
	}
}

func BenchmarkRead_SmallAggregate(b *testing.B) {
	runReadBenchmark(b, 5)
}

func BenchmarkRead_LargeAggregate(b *testing.B) {
	runReadBenchmark(b, 1000)
}

func runReadBenchmark(b *testing.B, eventsPerIP int) {
	dir, err := ioutil.TempDir("", "")
	if err != nil {
		b.Fatal(err)
	}
	defer os.RemoveAll(dir)

	db, err := Open(dir + "/test.db")
	if err != nil {
		b.Fatalf("Open: %v", err)
	}
	defer db.Close()

	// Write a lot of events just to bulk up the database, but
	// also write a couple of records for a known IP, that we're
	// going to read in the benchmark.
	n := manyEvents
	db.AddAggregate(randomEvents(n))

	refIP := "1.2.3.4"
	a := new(ippb.Aggregate)
	for i := 0; i < eventsPerIP; i++ {
		a.AddEvent(&ippb.Event{Type: "test", Ip: refIP, Count: 1})
	}
	db.AddAggregate(a)

	db.db.CompactRange(util.Range{})

	deadline := time.Now().Add(-1 * time.Hour)
	b.ResetTimer()

	for i := 0; i < b.N; i++ {
		m, err := db.ScanIP(deadline, refIP)
		if err != nil {
			b.Fatalf("ScanIP(%d): %v", i, err)
		}
		if len(m) == 0 {
			b.Fatalf("ScanIP(%d): returned empty result", i)
		}
		if len(m["test"]) < 1 {
			b.Fatalf("ScanIP(%d): returned bad results: %v", i, m)
		}
	}
}
